@@ -166,6 +166,17 @@ class InstantiationData(object):
         params = self.data.items()
         return dict((k, v) for k, v in params if k not in self.PARAMS)
 
+    def no_change(self):
+        assert self.patch
+        return ((self.template() is None) and
+                (self.environment() ==
+                    environment_format.default_for_missing({})) and
+                (not self.files()) and
+                (not self.environment_files()) and
+                (self.files_container() is None) and
+                (not any(k != rpc_api.PARAM_EXISTING
+                         for k in self.args().keys())))
+
 
 class StackController(object):
     """WSGI controller for stacks resource in Heat v1 API.
@@ -204,7 +215,7 @@ class StackController(object):
             raise exc.HTTPBadRequest(str(e))
 
     def _index(self, req, use_admin_cnxt=False):
-        filter_whitelist = {
+        filter_param_types = {
             # usage of keys in this list are not encouraged, please use
             # rpc_api.STACK_KEYS instead
             'id': util.PARAM_TYPE_MIXED,
@@ -215,7 +226,7 @@ class StackController(object):
             'username': util.PARAM_TYPE_MIXED,
             'owner_id': util.PARAM_TYPE_MIXED,
         }
-        whitelist = {
+        param_types = {
             'limit': util.PARAM_TYPE_SINGLE,
             'marker': util.PARAM_TYPE_SINGLE,
             'sort_dir': util.PARAM_TYPE_SINGLE,
@@ -228,7 +239,7 @@ class StackController(object):
             'not_tags': util.PARAM_TYPE_SINGLE,
             'not_tags_any': util.PARAM_TYPE_SINGLE,
         }
-        params = util.get_allowed_params(req.params, whitelist)
+        params = util.get_allowed_params(req.params, param_types)
         stack_keys = dict.fromkeys(rpc_api.STACK_KEYS, util.PARAM_TYPE_MIXED)
         unsupported = (
             rpc_api.STACK_ID,  # not user visible
@@ -246,7 +257,7 @@ class StackController(object):
         for key in unsupported:
             stack_keys.pop(key)
         # downward compatibility
-        stack_keys.update(filter_whitelist)
+        stack_keys.update(filter_param_types)
         filter_params = util.get_allowed_params(req.params, stack_keys)
 
         show_deleted = False
@@ -496,7 +507,8 @@ class StackController(object):
 
         raise exc.HTTPAccepted()
 
-    @util.registered_identified_stack
+    @util.no_policy_enforce
+    @util._identified_stack
     def update_patch(self, req, identity, body):
         """Update an existing stack with a new template.
 
@@ -504,6 +516,17 @@ class StackController(object):
         Add the flag patch to the args so the engine code can distinguish
         """
         data = InstantiationData(body, patch=True)
+        _target = {"project_id": req.context.tenant_id}
+
+        policy_act = 'update_no_change' if data.no_change() else 'update_patch'
+        allowed = req.context.policy.enforce(
+            context=req.context,
+            action=policy_act,
+            scope=self.REQUEST_SCOPE,
+            target=_target,
+            is_registered_policy=True)
+        if not allowed:
+            raise exc.HTTPForbidden()
 
         args = self.prepare_args(data, is_update=True)
         self.rpc_client.update_stack(
@@ -519,8 +542,8 @@ class StackController(object):
         raise exc.HTTPAccepted()
 
     def _param_show_nested(self, req):
-        whitelist = {'show_nested': 'single'}
-        params = util.get_allowed_params(req.params, whitelist)
+        param_types = {'show_nested': util.PARAM_TYPE_SINGLE}
+        params = util.get_allowed_params(req.params, param_types)
 
         p_name = 'show_nested'
         if p_name in params:
@@ -604,9 +627,9 @@ class StackController(object):
 
         data = InstantiationData(body)
 
-        whitelist = {'show_nested': util.PARAM_TYPE_SINGLE,
-                     'ignore_errors': util.PARAM_TYPE_SINGLE}
-        params = util.get_allowed_params(req.params, whitelist)
+        param_types = {'show_nested': util.PARAM_TYPE_SINGLE,
+                       'ignore_errors': util.PARAM_TYPE_SINGLE}
+        params = util.get_allowed_params(req.params, param_types)
 
         show_nested = False
         p_name = rpc_api.PARAM_SHOW_NESTED
