@@ -25,7 +25,6 @@ from oslo_utils import excutils
 from oslo_utils import timeutils as oslo_timeutils
 from oslo_utils import uuidutils
 from osprofiler import profiler
-import six
 
 from heat.common import context as common_context
 from heat.common import environment_format as env_fmt
@@ -75,20 +74,20 @@ class ForcedCancel(Exception):
 
 
 def reset_state_on_error(func):
-    @six.wraps(func)
+    @functools.wraps(func)
     def handle_exceptions(stack, *args, **kwargs):
         errmsg = None
         try:
             return func(stack, *args, **kwargs)
         except Exception as exc:
             with excutils.save_and_reraise_exception():
-                errmsg = six.text_type(exc)
+                errmsg = str(exc)
                 LOG.error('Unexpected exception in %(func)s: %(msg)s',
                           {'func': func.__name__, 'msg': errmsg})
         except BaseException as exc:
             with excutils.save_and_reraise_exception():
                 exc_type = type(exc).__name__
-                errmsg = '%s(%s)' % (exc_type, six.text_type(exc))
+                errmsg = '%s(%s)' % (exc_type, str(exc))
                 LOG.info('Stopped due to %(msg)s in %(func)s',
                          {'func': func.__name__, 'msg': errmsg})
         finally:
@@ -101,8 +100,7 @@ def reset_state_on_error(func):
     return handle_exceptions
 
 
-@six.python_2_unicode_compatible
-class Stack(collections.Mapping):
+class Stack(collections.abc.Mapping):
 
     ACTIONS = (
         CREATE, DELETE, UPDATE, ROLLBACK, SUSPEND, RESUME, ADOPT,
@@ -128,7 +126,7 @@ class Stack(collections.Mapping):
                  nested_depth=0, strict_validate=True, convergence=False,
                  current_traversal=None, tags=None, prev_raw_template_id=None,
                  current_deps=None, cache_data=None,
-                 deleted_time=None, converge=False):
+                 deleted_time=None, converge=False, refresh_cred=False):
 
         """Initialise the Stack.
 
@@ -190,6 +188,9 @@ class Stack(collections.Mapping):
         self.thread_group_mgr = None
         self.converge = converge
 
+        # This flag is use to check whether credential needs to refresh or not
+        self.refresh_cred = refresh_cred
+
         # strict_validate can be used to disable value validation
         # in the resource properties schema, this is useful when
         # performing validation when properties reference attributes
@@ -337,7 +338,7 @@ class Stack(collections.Mapping):
             resources = self._db_resources_get()
 
         stk_def_cache = {}
-        for rsc in six.itervalues(resources):
+        for rsc in resources.values():
             loaded_res = self._resource_from_db_resource(rsc, stk_def_cache)
             if loaded_res is not None:
                 yield loaded_res
@@ -520,14 +521,14 @@ class Stack(collections.Mapping):
         """
         if self._dependencies is None:
             deps = dependencies.Dependencies()
-            for res in six.itervalues(self.resources):
+            for res in self.resources.values():
                 res.add_explicit_dependencies(deps)
             self._dependencies = deps
         return self._dependencies
 
     def _add_implicit_dependencies(self, deps, ignore_errors=True):
         """Augment the given dependencies with implicit ones from plugins."""
-        for res in six.itervalues(self.resources):
+        for res in self.resources.values():
             try:
                 res.add_dependencies(deps)
             except Exception as exc:
@@ -540,13 +541,38 @@ class Stack(collections.Mapping):
                 else:
                     LOG.warning('Ignoring error adding implicit '
                                 'dependencies for %(res)s: %(err)s',
-                                {'res': six.text_type(res),
-                                 'err': six.text_type(exc)})
+                                {'res': str(res),
+                                 'err': str(exc)})
+
+    @classmethod
+    def _check_refresh_cred(cls, context, stack):
+        if stack.user_creds_id:
+            creds_obj = ucreds_object.UserCreds.get_by_id(
+                context, stack.user_creds_id)
+            creds = creds_obj.obj_to_primitive()["versioned_object.data"]
+            stored_context = common_context.StoredContext.from_dict(creds)
+
+            if cfg.CONF.deferred_auth_method == 'trusts':
+                old_trustor_proj_id = stored_context.tenant_id
+                old_trustor_user_id = stored_context.trustor_user_id
+
+                trustor_user_id = context.auth_plugin.get_user_id(
+                    context.clients.client('keystone').session)
+                trustor_proj_id = context.auth_plugin.get_project_id(
+                    context.clients.client('keystone').session)
+                return False if (
+                    old_trustor_user_id == trustor_user_id) and (
+                    old_trustor_proj_id == trustor_proj_id
+                    ) else True
+
+        # Should not raise error or allow refresh credential when we can't find
+        # user_creds_id in stack
+        return False
 
     @classmethod
     def load(cls, context, stack_id=None, stack=None, show_deleted=True,
              use_stored_context=False, force_reload=False, cache_data=None,
-             load_template=True):
+             load_template=True, check_refresh_cred=False):
         """Retrieve a Stack from the database."""
         if stack is None:
             stack = stack_object.Stack.get_by_id(
@@ -557,13 +583,22 @@ class Stack(collections.Mapping):
             message = _('No stack exists with id "%s"') % str(stack_id)
             raise exception.NotFound(message)
 
+        refresh_cred = False
+        if check_refresh_cred and (
+            cfg.CONF.deferred_auth_method == 'trusts'
+        ):
+            if cls._check_refresh_cred(context, stack):
+                use_stored_context = False
+                refresh_cred = True
+
         if force_reload:
             stack.refresh()
 
         return cls._from_db(context, stack,
                             use_stored_context=use_stored_context,
                             cache_data=cache_data,
-                            load_template=load_template)
+                            load_template=load_template,
+                            refresh_cred=refresh_cred)
 
     @classmethod
     def load_all(cls, context, limit=None, marker=None, sort_keys=None,
@@ -597,7 +632,7 @@ class Stack(collections.Mapping):
     @classmethod
     def _from_db(cls, context, stack,
                  use_stored_context=False, cache_data=None,
-                 load_template=True):
+                 load_template=True, refresh_cred=False):
         if load_template:
             template = tmpl.Template.load(
                 context, stack.raw_template_id, stack.raw_template)
@@ -621,7 +656,8 @@ class Stack(collections.Mapping):
                    prev_raw_template_id=stack.prev_raw_template_id,
                    current_deps=stack.current_deps, cache_data=cache_data,
                    nested_depth=stack.nested_depth,
-                   deleted_time=stack.deleted_at)
+                   deleted_time=stack.deleted_at,
+                   refresh_cred=refresh_cred)
 
     def get_kwargs_for_cloning(self, keep_status=False, only_db=False,
                                keep_tags=False):
@@ -655,7 +691,7 @@ class Stack(collections.Mapping):
             stack.update({
                 'action': self.action,
                 'status': self.status,
-                'status_reason': six.text_type(self.status_reason)})
+                'status_reason': str(self.status_reason)})
 
         if only_db:
             stack['parent_resource_name'] = self.parent_resource_name
@@ -689,6 +725,17 @@ class Stack(collections.Mapping):
             s['raw_template_id'] = self.t.id
 
         if self.id is not None:
+            if self.refresh_cred:
+                keystone = self.clients.client('keystone')
+                trust_ctx = keystone.regenerate_trust_context()
+                new_creds = ucreds_object.UserCreds.create(trust_ctx)
+                s['user_creds_id'] = new_creds.id
+
+                self._delete_user_cred(raise_keystone_exception=True)
+
+                self.user_creds_id = new_creds.id
+                self.refresh_cred = False
+
             if exp_trvsl is None and not ignore_traversal_check:
                 exp_trvsl = self.current_traversal
 
@@ -801,7 +848,7 @@ class Stack(collections.Mapping):
     def __str__(self):
         """Return a human-readable string representation of the stack."""
         text = 'Stack "%s" [%s]' % (self.name, self.id)
-        return six.text_type(text)
+        return str(text)
 
     def resource_by_refid(self, refid):
         """Return the resource in this stack with the specified refid.
@@ -809,7 +856,7 @@ class Stack(collections.Mapping):
         :returns: resource in this stack with the specified refid, or None if
                   not found.
         """
-        for r in six.itervalues(self):
+        for r in self.values():
             if r.state not in ((r.INIT, r.COMPLETE),
                                (r.CREATE, r.IN_PROGRESS),
                                (r.CREATE, r.COMPLETE),
@@ -906,7 +953,7 @@ class Stack(collections.Mapping):
         else:
             iter_rsc = self._explicit_dependencies()
 
-        unique_defns = set(res.t for res in six.itervalues(resources))
+        unique_defns = set(res.t for res in resources.values())
         unique_defn_names = set(defn.name for defn in unique_defns)
 
         for res in iter_rsc:
@@ -939,7 +986,7 @@ class Stack(collections.Mapping):
                 raise exception.StackValidationFailed(message=result)
             eventlet.sleep(0)
 
-        for op_name, output in six.iteritems(self.outputs):
+        for op_name, output in self.outputs.items():
             try:
                 output.validate()
             except exception.StackValidationFailed as ex:
@@ -958,7 +1005,7 @@ class Stack(collections.Mapping):
         during its lifecycle using the configured deferred authentication
         method.
         """
-        return any(res.requires_deferred_auth for res in six.itervalues(self))
+        return any(res.requires_deferred_auth for res in self.values())
 
     def _add_event(self, action, status, reason):
         """Add a state change event to the database."""
@@ -1045,7 +1092,7 @@ class Stack(collections.Mapping):
         if stack is not None:
             values = {'action': self.action,
                       'status': self.status,
-                      'status_reason': six.text_type(self.status_reason)}
+                      'status_reason': str(self.status_reason)}
             self._send_notification_and_add_event()
             if self.convergence:
                 # do things differently for convergence
@@ -1075,7 +1122,7 @@ class Stack(collections.Mapping):
         if stack is not None:
             values = {'action': self.action,
                       'status': self.status,
-                      'status_reason': six.text_type(self.status_reason)}
+                      'status_reason': str(self.status_reason)}
             self._send_notification_and_add_event()
             stack.persist_state_and_release_lock(self.context, self.id,
                                                  engine_id, values)
@@ -1095,7 +1142,7 @@ class Stack(collections.Mapping):
     def preview_resources(self):
         """Preview the stack with all of the resources."""
         return [resource.preview()
-                for resource in six.itervalues(self.resources)]
+                for resource in self.resources.values()]
 
     def get_nested_parameters(self, filter_func):
         """Return nested parameters schema, if any.
@@ -1105,7 +1152,7 @@ class Stack(collections.Mapping):
         stack.
         """
         result = {}
-        for name, rsrc in six.iteritems(self.resources):
+        for name, rsrc in self.resources.items():
             nested = rsrc.get_nested_parameters_stack()
             if nested is None:
                 continue
@@ -1157,7 +1204,6 @@ class Stack(collections.Mapping):
 
         return {'resource_data': data['resources'].get(resource.name)}
 
-    @scheduler.wrappertask
     def stack_task(self, action, reverse=False, post_func=None,
                    aggregate_exceptions=False, pre_completion_func=None,
                    notify=None):
@@ -1182,7 +1228,7 @@ class Stack(collections.Mapping):
                                               None, action)
         except Exception as e:
             self.state_set(action, self.FAILED, e.args[0] if e.args else
-                           'Failed stack pre-ops: %s' % six.text_type(e))
+                           'Failed stack pre-ops: %s' % str(e))
             if callable(post_func):
                 post_func()
             if notify is not None:
@@ -1206,12 +1252,11 @@ class Stack(collections.Mapping):
                                 lambda x: {})
 
         @functools.wraps(getattr(resource.Resource, action_method))
-        @scheduler.wrappertask
         def resource_action(r):
             # Find e.g resource.create and call it
             handle = getattr(r, action_method)
 
-            yield handle(**handle_kwargs(r))
+            yield from handle(**handle_kwargs(r))
 
             if action == self.CREATE:
                 stk_defn.update_resource_data(self.defn, r.name, r.node_data())
@@ -1227,7 +1272,7 @@ class Stack(collections.Mapping):
             aggregate_exceptions=aggregate_exceptions)
 
         try:
-            yield action_task()
+            yield from action_task()
         except scheduler.Timeout:
             stack_status = self.FAILED
             reason = '%s timed out' % action.title()
@@ -1238,7 +1283,7 @@ class Stack(collections.Mapping):
             # ExceptionGroup, but the raw exception.
             # see scheduler.py line 395-399
             stack_status = self.FAILED
-            reason = 'Resource %s failed: %s' % (action, six.text_type(ex))
+            reason = 'Resource %s failed: %s' % (action, str(ex))
 
         if pre_completion_func:
             pre_completion_func(self, action, stack_status, reason)
@@ -1272,7 +1317,7 @@ class Stack(collections.Mapping):
                 return hasattr(res, 'handle_%s' % res.CHECK.lower())
 
         all_supported = all(is_supported(res)
-                            for res in six.itervalues(self.resources))
+                            for res in self.resources.values())
 
         if not all_supported:
             msg = ". '%s' not fully supported (see resources)" % self.CHECK
@@ -1316,7 +1361,7 @@ class Stack(collections.Mapping):
             if not self.disable_rollback and self.state == (self.ADOPT,
                                                             self.FAILED):
                 # enter the same flow as abandon and just delete the stack
-                for res in six.itervalues(self.resources):
+                for res in self.resources.values():
                     res.abandon_in_progress = True
                 self.delete(action=self.ROLLBACK, abandon=True)
 
@@ -1426,7 +1471,7 @@ class Stack(collections.Mapping):
             try:
                 self.delete_all_snapshots()
             except Exception as exc:
-                self.state_set(self.action, self.FAILED, six.text_type(exc))
+                self.state_set(self.action, self.FAILED, str(exc))
                 self.purge_db()
                 return
 
@@ -1499,7 +1544,10 @@ class Stack(collections.Mapping):
                     # Rolling back to previous resource
                     score += 10
 
-                return score, ext_rsrc.updated_at
+                last_changed_at = ext_rsrc.updated_at
+                if last_changed_at is None:
+                    last_changed_at = ext_rsrc.created_at
+                return score, last_changed_at
 
             candidates = sorted((r for r in self.ext_rsrcs_db.values()
                                  if r.name == rsrc_name),
@@ -1576,11 +1624,11 @@ class Stack(collections.Mapping):
         return set(n.rsrc_id for n in dep_nodes if not n.is_update)
 
     def reset_stack_and_resources_in_progress(self, reason):
-        for name, rsrc in six.iteritems(self.resources):
+        for name, rsrc in self.resources.items():
             if rsrc.status == rsrc.IN_PROGRESS:
                 rsrc.state_set(rsrc.action,
                                rsrc.FAILED,
-                               six.text_type(reason))
+                               str(reason))
         if self.action == self.UPDATE and not self.convergence:
             backup_stack = self._backup_stack(False)
             existing_params = environment.Environment({env_fmt.PARAMETERS:
@@ -1591,9 +1639,8 @@ class Stack(collections.Mapping):
             self._merge_user_param_template(existing_params, template,
                                             bkp_stack_template)
 
-        self.state_set(self.action, self.FAILED, six.text_type(reason))
+        self.state_set(self.action, self.FAILED, str(reason))
 
-    @scheduler.wrappertask
     def update_task(self, newstack, action=UPDATE,
                     msg_queue=None, notify=None):
         if action not in (self.UPDATE, self.ROLLBACK, self.RESTORE):
@@ -1609,7 +1656,7 @@ class Stack(collections.Mapping):
                                               newstack, action)
         except Exception as e:
             self.state_set(action, self.FAILED, e.args[0] if e.args else
-                           'Failed stack pre-ops: %s' % six.text_type(e))
+                           'Failed stack pre-ops: %s' % str(e))
             if notify is not None:
                 notify.signal()
             return
@@ -1676,8 +1723,8 @@ class Stack(collections.Mapping):
             check_message = functools.partial(self._check_for_message,
                                               msg_queue)
             try:
-                yield updater.as_task(timeout=self.timeout_secs(),
-                                      progress_callback=check_message)
+                yield from updater.as_task(timeout=self.timeout_secs(),
+                                           progress_callback=check_message)
             finally:
                 self.reset_dependencies()
 
@@ -1693,7 +1740,7 @@ class Stack(collections.Mapping):
             # so we roll back to the original state
             should_rollback = self._update_exception_handler(e, action)
             if should_rollback:
-                yield self.update_task(oldstack, action=self.ROLLBACK)
+                yield from self.update_task(oldstack, action=self.ROLLBACK)
         except BaseException as e:
             with excutils.save_and_reraise_exception():
                 self._update_exception_handler(e, action)
@@ -1756,7 +1803,7 @@ class Stack(collections.Mapping):
 
         :returns: a boolean for require rollback flag.
         """
-        self.status_reason = six.text_type(exc)
+        self.status_reason = str(exc)
         self.status = self.FAILED
         if action != self.UPDATE:
             return False
@@ -1800,7 +1847,7 @@ class Stack(collections.Mapping):
 
         def copy_data(source_res, destination_res):
             if source_res.data():
-                for key, val in six.iteritems(source_res.data()):
+                for key, val in source_res.data().items():
                     destination_res.data_set(key, val)
 
         for key, backup_res in stack.resources.items():
@@ -1845,11 +1892,10 @@ class Stack(collections.Mapping):
             LOG.exception("Failed to retrieve user_creds")
             return None
 
-    def _delete_credentials(self, stack_status, reason, abandon):
+    def _delete_user_cred(self, stack_status=None, reason=None,
+                          raise_keystone_exception=False):
         # Cleanup stored user_creds so they aren't accessible via
         # the soft-deleted stack which remains in the DB
-        # The stack_status and reason passed in are current values, which
-        # may get rewritten and returned from this method
         if self.user_creds_id:
             user_creds = self._try_get_user_creds()
             # If we created a trust, delete it
@@ -1879,6 +1925,8 @@ class Stack(collections.Mapping):
                         # Without this, they would need to issue
                         # an additional stack-delete
                         LOG.exception("Error deleting trust")
+                        if raise_keystone_exception:
+                            raise
 
             # Delete the stored credentials
             try:
@@ -1888,13 +1936,18 @@ class Stack(collections.Mapping):
                 LOG.info("Tried to delete user_creds that do not exist "
                          "(stack=%(stack)s user_creds_id=%(uc)s)",
                          {'stack': self.id, 'uc': self.user_creds_id})
+            self.user_creds_id = None
+        return stack_status, reason
 
-            try:
-                self.user_creds_id = None
-                self.store()
-            except exception.NotFound:
-                LOG.info("Tried to store a stack that does not exist %s",
-                         self.id)
+    def _delete_credentials(self, stack_status, reason, abandon):
+        # The stack_status and reason passed in are current values, which
+        # may get rewritten and returned from this method
+        stack_status, reason = self._delete_user_cred(stack_status, reason)
+        try:
+            self.store()
+        except exception.NotFound:
+            LOG.info("Tried to store a stack that does not exist %s",
+                     self.id)
 
         # If the stack has a domain project, delete it
         if self.stack_user_project_id and not abandon:
@@ -1905,7 +1958,7 @@ class Stack(collections.Mapping):
             except Exception as ex:
                 LOG.exception("Error deleting project")
                 stack_status = self.FAILED
-                reason = "Error deleting project: %s" % six.text_type(ex)
+                reason = "Error deleting project: %s" % str(ex)
 
         return stack_status, reason
 
@@ -1933,8 +1986,8 @@ class Stack(collections.Mapping):
 
         stack_status = self.COMPLETE
         reason = 'Stack %s completed successfully' % action
-        self.state_set(action, self.IN_PROGRESS, 'Stack %s started' %
-                       action)
+        self.state_set(action, self.IN_PROGRESS, 'Stack %s started at %s' %
+                       (action, oslo_timeutils.utcnow().isoformat()))
         if notify is not None:
             notify.signal()
 
@@ -1957,7 +2010,7 @@ class Stack(collections.Mapping):
             except Exception as e:
                 self.state_set(action, self.FAILED,
                                e.args[0] if e.args else
-                               'Failed stack pre-ops: %s' % six.text_type(e))
+                               'Failed stack pre-ops: %s' % str(e))
                 return
 
         action_task = scheduler.DependencyTaskGroup(self.dependencies,
@@ -1967,7 +2020,7 @@ class Stack(collections.Mapping):
             scheduler.TaskRunner(action_task)(timeout=self.timeout_secs())
         except exception.ResourceFailure as ex:
             stack_status = self.FAILED
-            reason = 'Resource %s failed: %s' % (action, six.text_type(ex))
+            reason = 'Resource %s failed: %s' % (action, str(ex))
         except scheduler.Timeout:
             stack_status = self.FAILED
             reason = '%s timed out' % action.title()
@@ -2092,7 +2145,7 @@ class Stack(collections.Mapping):
             ss_defn = self.defn.clone_with_new_template(template,
                                                         self.identifier())
             resources = self._resources_for_defn(ss_defn)
-            for name, rsrc in six.iteritems(resources):
+            for name, rsrc in resources.items():
                 data = snapshot.data['resources'].get(name)
                 if data:
                     scheduler.TaskRunner(rsrc.delete_snapshot, data)()
@@ -2159,7 +2212,7 @@ class Stack(collections.Mapping):
             'status': self.status,
             'template': self.t.t,
             'resources': dict((res.name, res.prepare_abandon())
-                              for res in six.itervalues(self.resources)),
+                              for res in self.resources.values()),
             'project_id': self.tenant_id,
             'stack_user_project_id': self.stack_user_project_id,
             'tags': self.tags,
@@ -2279,9 +2332,9 @@ class Stack(collections.Mapping):
                 requires = set(res_id_dep.requires(db_res.id))
                 r = self.resources.get(db_res.name)
                 if r is None:
-                    # delete db resources not in current_template_id
+                    # delete DB resources not in current_template_id
                     LOG.warning("Resource %(res)s not found in template "
-                                "for stack %(st)s, deleting from db.",
+                                "for stack %(st)s, deleting from DB.",
                                 {'res': db_res.name, 'st': self.id})
                     resource_objects.Resource.delete(self.context, db_res.id)
                 else:
