@@ -11,8 +11,7 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-import mock
-import six
+from unittest import mock
 
 from glanceclient import exc
 from heat.common import exception
@@ -112,11 +111,12 @@ class GlanceImageTest(common.HeatTestCase):
         glance.return_value = self.glanceclient
         self.images = self.glanceclient.images
         self.image_tags = self.glanceclient.image_tags
+        self.image_members = self.glanceclient.image_members
 
     def _test_validate(self, resource, error_msg):
         exc = self.assertRaises(exception.StackValidationFailed,
                                 resource.validate)
-        self.assertIn(error_msg, six.text_type(exc))
+        self.assertIn(error_msg, str(exc))
 
     def test_invalid_min_disk(self):
         # invalid 'min_disk'
@@ -478,11 +478,12 @@ class GlanceWebImageTest(common.HeatTestCase):
         glance.return_value = self.glanceclient
         self.images = self.glanceclient.images
         self.image_tags = self.glanceclient.image_tags
+        self.image_members = self.glanceclient.image_members
 
     def _test_validate(self, resource, error_msg):
         exc = self.assertRaises(exception.StackValidationFailed,
                                 resource.validate)
-        self.assertIn(error_msg, six.text_type(exc))
+        self.assertIn(error_msg, str(exc))
 
     def test_invalid_min_disk(self):
         # invalid 'min_disk'
@@ -641,9 +642,49 @@ class GlanceWebImageTest(common.HeatTestCase):
             name=u'cirros_image',
             protected=False,
             owner=u'test_owner',
-            tags=['tag1'],
+            tags=['tag1']
         )
 
+    def test_image_active_property_image_not_active(self):
+        self.images.reactivate.return_value = None
+        self.images.deactivate.return_value = None
+        value = mock.MagicMock()
+        image_id = '41f0e60c-ebb4-4375-a2b4-845ae8b9c995'
+        value.id = image_id
+        value.status = 'pending'
+        self.images.create.return_value = value
+        self.my_image.handle_create()
+        self.my_image.check_create_complete(image_id)
+        self.images.deactivate.assert_not_called()
+
+    def test_image_active_property_image_active_to_deactivate(self):
+        self.images.reactivate.return_value = None
+        self.images.deactivate.return_value = None
+        value = mock.MagicMock()
+        image_id = '41f0e60c-ebb4-4375-a2b4-845ae8b9c995'
+        value.id = image_id
+        value.status = 'active'
+        self.my_image.resource_id = image_id
+        self.images.create.return_value = value
+        self.images.get.return_value = value
+        self.my_image.check_create_complete(False)
+        self.images.deactivate.assert_called_once_with(
+            self.my_image.resource_id)
+
+    def test_image_active_property_image_status_killed(self):
+        self.images.reactivate.return_value = None
+        self.images.deactivate.return_value = None
+        value = mock.MagicMock()
+        image_id = '41f0e60c-ebb4-4375-a2b4-845ae8b9c995'
+        value.id = image_id
+        value.status = 'killed'
+        self.my_image.resource_id = image_id
+        self.images.create.return_value = value
+        self.images.get.return_value = value
+        ex = self.assertRaises(exception.ResourceInError,
+                               self.my_image.check_create_complete, False)
+        self.assertIn('killed', ex.message)
+
     def _handle_update_tags(self, prop_diff):
         self.my_image.handle_update(json_snippet=None,
                                     tmpl_diff=None,
@@ -679,6 +720,49 @@ class GlanceWebImageTest(common.HeatTestCase):
             ramdisk_id='12345678-1234-1234-1234-123456789012'
         )
 
+    def test_image_handle_update_deactivate(self):
+        self.images.reactivate.return_value = None
+        self.images.deactivate.return_value = None
+        value = mock.MagicMock()
+        image_id = '41f0e60c-ebb4-4375-a2b4-845ae8b9c995'
+        value.id = image_id
+        value.status = 'active'
+        self.my_image.resource_id = image_id
+        props = self.stack.t.t['resources']['my_image']['properties'].copy()
+        props['active'] = False
+        self.my_image.t = self.my_image.t.freeze(properties=props)
+        prop_diff = {'active': False}
+        self.my_image.reparse()
+        self.images.update.return_value = value
+        self.images.get.return_value = value
+        self.my_image.handle_update(json_snippet=None,
+                                    tmpl_diff=None,
+                                    prop_diff=prop_diff)
+        self.images.deactivate.assert_called_once_with(
+            self.my_image.resource_id)
+
+    def test_image_handle_update_reactivate(self):
+        self.images.reactivate.return_value = None
+        self.images.deactivate.return_value = None
+        value = mock.MagicMock()
+        image_id = '41f0e60c-ebb4-4375-a2b4-845ae8b9c995'
+        value.id = image_id
+        value.status = 'deactivated'
+        self.my_image.resource_id = image_id
+        props = self.stack.t.t['resources']['my_image']['properties'].copy()
+        props['active'] = True
+        self.my_image.t = self.my_image.t.freeze(properties=props)
+        prop_diff = {'active': True}
+        self.my_image.reparse()
+        self.images.update.return_value = value
+        self.images.get.return_value = value
+        self.my_image.handle_update(json_snippet=None,
+                                    tmpl_diff=None,
+                                    prop_diff=prop_diff)
+        self.my_image.check_update_complete(True)
+        self.images.reactivate.assert_called_once_with(
+            self.my_image.resource_id)
+
     def test_image_handle_update_tags(self):
         self.my_image.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
 
@@ -708,6 +792,49 @@ class GlanceWebImageTest(common.HeatTestCase):
             'tag1'
         )
 
+    def _handle_update_members(self, prop_diff):
+        self.my_image.handle_update(json_snippet=None,
+                                    tmpl_diff=None,
+                                    prop_diff=prop_diff)
+
+        self.image_members.create.assert_called_once_with(
+            self.my_image.resource_id,
+            'member2'
+        )
+        self.image_members.delete.assert_called_once_with(
+            self.my_image.resource_id,
+            'member1'
+        )
+
+    def test_image_handle_update_members(self):
+        self.my_image.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
+
+        props = self.stack.t.t['resources']['my_image']['properties'].copy()
+        props['members'] = ['member1']
+        self.my_image.t = self.my_image.t.freeze(properties=props)
+        self.my_image.reparse()
+        prop_diff = {'members': ['member2']}
+
+        self._handle_update_members(prop_diff)
+
+    def test_image_handle_update_remove_members(self):
+        self.my_image.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
+
+        props = self.stack.t.t['resources']['my_image']['properties'].copy()
+        props['members'] = ['member1']
+        self.my_image.t = self.my_image.t.freeze(properties=props)
+        self.my_image.reparse()
+        prop_diff = {'members': None}
+
+        self.my_image.handle_update(json_snippet=None,
+                                    tmpl_diff=None,
+                                    prop_diff=prop_diff)
+
+        self.image_members.delete.assert_called_once_with(
+            self.my_image.resource_id,
+            'member1'
+        )
+
     def test_image_handle_update_tags_delete_not_found(self):
         self.my_image.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
 
@@ -751,6 +878,7 @@ class GlanceWebImageTest(common.HeatTestCase):
             'name': 'test',
             'disk_format': 'qcow2',
             'container_format': 'bare',
+            'active': None,
             'protected': False,
             'is_public': False,
             'min_disk': 0,
@@ -763,6 +891,7 @@ class GlanceWebImageTest(common.HeatTestCase):
             'os_version': '1.0',
             'owner': 'test_owner',
             'ramdisk_id': '12345678-1234-1234-1234-123456789012',
+            'members': None,
             'visibility': 'private'
         }
         image = show_value
@@ -774,6 +903,7 @@ class GlanceWebImageTest(common.HeatTestCase):
             'name': 'test',
             'disk_format': 'qcow2',
             'container_format': 'bare',
+            'active': None,
             'protected': False,
             'min_disk': 0,
             'min_ram': 0,
@@ -785,6 +915,7 @@ class GlanceWebImageTest(common.HeatTestCase):
             'os_version': '1.0',
             'owner': 'test_owner',
             'ramdisk_id': '12345678-1234-1234-1234-123456789012',
+            'members': None,
             'visibility': 'private'
         }
 
