@@ -13,11 +13,10 @@
 
 import os.path
 import sys
+from unittest import mock
 
 import fixtures
-import mock
 from oslo_config import cfg
-import six
 
 from heat.common import environment_format
 from heat.common import exception
@@ -209,7 +208,7 @@ def constraint_mapping():
         env = environment.Environment({})
         error = self.assertRaises(ValueError,
                                   resources._load_global_environment, env)
-        self.assertEqual("oops", six.text_type(error))
+        self.assertEqual("oops", str(error))
 
     def test_constraints_registry_stevedore(self):
         env = environment.Environment({})
@@ -788,7 +787,7 @@ class ResourceRegistryTest(common.HeatTestCase):
                '\'post-delete\')')
         ex = self.assertRaises(exception.InvalidBreakPointHook,
                                registry.load, {'resources': resources})
-        self.assertEqual(msg, six.text_type(ex))
+        self.assertEqual(msg, str(ex))
 
     def test_list_type_validation_invalid_support_status(self):
         registry = environment.ResourceRegistry(None, {})
@@ -797,7 +796,7 @@ class ResourceRegistryTest(common.HeatTestCase):
                                registry.get_types,
                                support_status='junk')
         msg = ('Invalid support status and should be one of %s' %
-               six.text_type(support.SUPPORT_STATUSES))
+               str(support.SUPPORT_STATUSES))
 
         self.assertIn(msg, ex.message)
 
