@@ -11,12 +11,13 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-import mock
+from unittest import mock
+
 from oslo_serialization import jsonutils
-import six
 
 from heat.common import exception
 from heat.engine import constraints
+from heat.engine import function
 from heat.engine.hot import functions as hot_funcs
 from heat.engine.hot import parameters as hot_param
 from heat.engine import parameters
@@ -142,9 +143,9 @@ class PropertySchemaTest(common.HeatTestCase):
 
     def test_all_resource_schemata(self):
         for resource_type in resources.global_env().get_types():
-            for schema in six.itervalues(getattr(resource_type,
-                                                 'properties_schema',
-                                                 {})):
+            for schema in getattr(resource_type,
+                                  'properties_schema',
+                                  {}).values():
                 properties.Schema.from_legacy(schema)
 
     def test_from_legacy_idempotency(self):
@@ -636,7 +637,7 @@ class PropertySchemaTest(common.HeatTestCase):
                 update = True
             sub_schema = prop.schema
             if sub_schema:
-                for sub_prop_key, sub_prop in six.iteritems(sub_schema):
+                for sub_prop_key, sub_prop in sub_schema.items():
                     if not update:
                         self.assertEqual(update, sub_prop.update_allowed,
                                          "Mismatch in update policies: "
@@ -652,11 +653,11 @@ class PropertySchemaTest(common.HeatTestCase):
                         check_update_policy(resource_type, sub_prop_key,
                                             sub_prop, update)
 
-        for resource_type, resource_class in six.iteritems(all_resources):
+        for resource_type, resource_class in all_resources.items():
             props_schemata = properties.schemata(
                 resource_class.properties_schema)
 
-            for prop_key, prop in six.iteritems(props_schemata):
+            for prop_key, prop in props_schemata.items():
                 check_update_policy(resource_type, prop_key, prop)
 
 
@@ -808,7 +809,7 @@ class PropertyTest(common.HeatTestCase):
         schema = {'Type': 'Integer'}
         p = properties.Property(schema)
         ex = self.assertRaises(TypeError, p.get_value, '3a')
-        self.assertEqual("Value '3a' is not an integer", six.text_type(ex))
+        self.assertEqual("Value '3a' is not an integer", str(ex))
 
     def test_integer_low(self):
         schema = {'Type': 'Integer',
@@ -1005,7 +1006,7 @@ class PropertyTest(common.HeatTestCase):
         ex = self.assertRaises(exception.StackValidationFailed,
                                p.get_value, {'valid': 'fish'}, True)
         self.assertEqual('Property error: valid: "fish" is not a '
-                         'valid boolean', six.text_type(ex))
+                         'valid boolean', str(ex))
 
     def test_map_schema_missing_data(self):
         map_schema = {'valid': {'Type': 'Boolean'}}
@@ -1018,7 +1019,7 @@ class PropertyTest(common.HeatTestCase):
         ex = self.assertRaises(exception.StackValidationFailed,
                                p.get_value, {}, True)
         self.assertEqual('Property error: Property valid not assigned',
-                         six.text_type(ex))
+                         str(ex))
 
     def test_list_schema_good(self):
         map_schema = {'valid': {'Type': 'Boolean'}}
@@ -1037,7 +1038,7 @@ class PropertyTest(common.HeatTestCase):
                                p.get_value,
                                [{'valid': 'True'}, {'valid': 'fish'}], True)
         self.assertEqual('Property error: [1].valid: "fish" is not '
-                         'a valid boolean', six.text_type(ex))
+                         'a valid boolean', str(ex))
 
     def test_list_schema_int_good(self):
         list_schema = {'Type': 'Integer'}
@@ -1050,7 +1051,7 @@ class PropertyTest(common.HeatTestCase):
         ex = self.assertRaises(exception.StackValidationFailed,
                                p.get_value, [42, 'fish'], True)
         self.assertEqual("Property error: [1]: Value 'fish' is not "
-                         "an integer", six.text_type(ex))
+                         "an integer", str(ex))
 
 
 class PropertiesTest(common.HeatTestCase):
@@ -1073,7 +1074,7 @@ class PropertiesTest(common.HeatTestCase):
             'default_override': 21,
         }
 
-        def double(d):
+        def double(d, nullable=False):
             return d * 2
 
         self.props = properties.Properties(schema, data, double, 'wibble')
@@ -1118,7 +1119,7 @@ class PropertiesTest(common.HeatTestCase):
         ex = self.assertRaises(KeyError, self.props.get_user_value, 'foo')
         # Note we have to use args here: https://bugs.python.org/issue2651
         self.assertEqual('Invalid Property foo',
-                         six.text_type(ex.args[0]))
+                         str(ex.args[0]))
 
     def test_bad_key(self):
         self.assertEqual('wibble', self.props.get('foo', 'wibble'))
@@ -1127,7 +1128,7 @@ class PropertiesTest(common.HeatTestCase):
         ex = self.assertRaises(KeyError, self.props.__getitem__, 'foo')
         # Note we have to use args here: https://bugs.python.org/issue2651
         self.assertEqual('Invalid Property foo',
-                         six.text_type(ex.args[0]))
+                         str(ex.args[0]))
 
     def test_none_string(self):
         schema = {'foo': {'Type': 'String'}}
@@ -1208,7 +1209,7 @@ class PropertiesTest(common.HeatTestCase):
     def test_resolve_returns_none(self):
         schema = {'foo': {'Type': 'String', "MinLength": "5"}}
 
-        def test_resolver(prop):
+        def test_resolver(prop, nullable=False):
             return None
 
         self.patchobject(properties.Properties,
@@ -1245,7 +1246,7 @@ class PropertiesTest(common.HeatTestCase):
         }
 
         # define parameters for function
-        def test_resolver(prop):
+        def test_resolver(prop, nullable=False):
             return 'None'
 
         class rsrc(object):
@@ -1676,6 +1677,26 @@ class PropertiesTest(common.HeatTestCase):
         bar_props = bar_rsrc.properties(schema)
         self.assertEqual('bar', bar_props['description'])
 
+    def test_null_property_value(self):
+        class NullFunction(function.Function):
+            def result(self):
+                return Ellipsis
+
+        schema = {
+            'Foo': properties.Schema('String', required=False),
+            'Bar': properties.Schema('String', required=False),
+            'Baz': properties.Schema('String', required=False),
+        }
+        user_props = {'Foo': NullFunction(None, 'null', []), 'Baz': None}
+        props = properties.Properties(schema, user_props, function.resolve)
+
+        self.assertEqual(None, props['Foo'])
+        self.assertEqual(None, props.get_user_value('Foo'))
+        self.assertEqual(None, props['Bar'])
+        self.assertEqual(None, props.get_user_value('Bar'))
+        self.assertEqual('', props['Baz'])
+        self.assertEqual('', props.get_user_value('Baz'))
+
 
 class PropertiesValidationTest(common.HeatTestCase):
     def test_required(self):
@@ -1713,14 +1734,14 @@ class PropertiesValidationTest(common.HeatTestCase):
         props = properties.Properties(schema, {'foo': ['foo', 'bar']})
         ex = self.assertRaises(exception.StackValidationFailed, props.validate)
         self.assertIn('Property error: foo: Value must be a string',
-                      six.text_type(ex))
+                      str(ex))
 
     def test_dict_instead_string(self):
         schema = {'foo': {'Type': 'String'}}
         props = properties.Properties(schema, {'foo': {'foo': 'bar'}})
         ex = self.assertRaises(exception.StackValidationFailed, props.validate)
         self.assertIn('Property error: foo: Value must be a string',
-                      six.text_type(ex))
+                      str(ex))
 
     def test_none_string(self):
         schema = {'foo': {'Type': 'String'}}
@@ -1887,7 +1908,7 @@ class PropertiesValidationTest(common.HeatTestCase):
         ex = self.assertRaises(exception.StackValidationFailed,
                                props.validate)
         self.assertEqual('Property error: foo[0]: Unknown Property bar',
-                         six.text_type(ex))
+                         str(ex))
 
     def test_nested_properties_schema_invalid_property_in_map(self):
         child_schema = {'Key': {'Type': 'String',
@@ -1906,7 +1927,7 @@ class PropertiesValidationTest(common.HeatTestCase):
         ex = self.assertRaises(exception.StackValidationFailed,
                                props.validate)
         self.assertEqual('Property error: foo.boo: Unknown Property bar',
-                         six.text_type(ex))
+                         str(ex))
 
     def test_more_nested_properties_schema_invalid_property_in_list(self):
         nested_child_schema = {'Key': {'Type': 'String',
@@ -1924,7 +1945,7 @@ class PropertiesValidationTest(common.HeatTestCase):
         ex = self.assertRaises(exception.StackValidationFailed,
                                props.validate)
         self.assertEqual('Property error: foo[0].doo: Unknown Property bar',
-                         six.text_type(ex))
+                         str(ex))
 
     def test_more_nested_properties_schema_invalid_property_in_map(self):
         nested_child_schema = {'Key': {'Type': 'String',
@@ -1942,7 +1963,7 @@ class PropertiesValidationTest(common.HeatTestCase):
         ex = self.assertRaises(exception.StackValidationFailed,
                                props.validate)
         self.assertEqual('Property error: foo.boo.doo: Unknown Property bar',
-                         six.text_type(ex))
+                         str(ex))
 
     def test_schema_to_template_empty_schema(self):
         schema = {}
